/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 931 $ $LastChangedDate: 2002-09-11 10:32:30 +0200 (Wed, 11 Sep 2002) $
 *
 * Ciface - dump .h file for given Gont module.
 */
 
open Ast;
open Util;
open Module;
open Tyutil;

string ctype(ty t)
{
    switch t {
    case Int:		return "gont_int";
    case Bool:		return "gont_int";
    case Void:		return "void";
    case Float:		return "double *";
    case String:	return "struct gont_string *";
    case Tyvar(_):	return "void *";
    case Abstract_type:	return "void *";
    case Exn:		return "struct gont_union *";
    case Tuple(_):	return "void **";
    case Func(_):	return "struct gont_lambda_expr *";
    case Struct(_):	return "void *";
    case Opt_struct(_):	return "void *";
    case Union(_):	return "struct gont_union *";
    case Named_type(_):	return ctype(type_instance(t));
    case X_tyvar(x):
	switch x.t {
	case Some(tt):	return ctype(tt);
	case None:	return "void*";
	}
    }
}

string sctype(ty t)
{
    switch t {
    case Int:		return "gont_int";
    case Bool:		return "char";
    case Void:		return "void";
    case Float:		return "double";
    case String:	return "struct gont_string *";
    case Tyvar(_):	return "void *";
    case Abstract_type:	return "void *";
    case Exn:		return "struct gont_union *";
    case Tuple(_):	return "void **";
    case Func(_):	return "struct gont_lambda_expr *";
    case Struct(_):	return "void *";
    case Opt_struct(_):	return "void *";
    case Union(_):	return "struct gont_union *";
    case Named_type(_):	return sctype(type_instance(t));
    case X_tyvar(x):
	switch x.t {
	case Some(tt):	return sctype(tt);
	case None:	return "void*";
	}
    }
}

void dump_ciface(id mod)
{
    void dump_var(string, var_symbol s)
    {
	if (has_property(s, Is_exception)) {
	    puts("extern gont_int " + s.a.extern_name + "; /* exn */\n");
	} else if (has_property(s, Is_union_field)) {
	    skip;
	} else if (is_global_function(s)) {
	    let ft = get_ft_record(s.stype) in
	    let parms = List::map(ctype, List::map(fst, ft.parms)) in
	    puts(ctype(ft.ret_type) + " " + s.a.extern_name + "(" +
		 String::concat(", ", parms) + ");\n");
	} else {
	    puts("extern " + ctype(s.stype) + " " + s.a.extern_name + 
		 "; /* value */\n");
	}
    }

    void dump_tysym(string, ty_symbol s)
    {
	int id = 1;

	void dump_ufld(<var_symbol>field f)
	{
	    let sn = (get_binding(f.fld_symbol)).a.extern_name in
	    let name = String::concat("__", String::split("$$", sn)) in
	    puts("#define GontUF_" + name + " " + itoa(id++) + "\n");
	}
	
	void dump_struct(<<sf_symbol>field>list flds)
	{
	    void dump_sfld(<sf_symbol>field f)
	    {
		let sn = (get_binding(f.fld_symbol)).name in
		puts("\t" + sctype(f.fld_type) + " " + sn + ";\n");
	    }

	    puts("struct GontS_" + s.module + "__" + s.name + " {\n");
	    List::iter(dump_sfld, flds);
	    puts("};\n\n");
	}

	puts("/* type " + s.name + " */\n");
	
	switch s.stype {
	case Union(flds):
	    List::iter(dump_ufld, flds);
	    
	case Struct(flds):
	    dump_struct(flds);
	    
	case Opt_struct(flds):
	    dump_struct(flds);
	    
	case _:
	    skip;
	}
    }

    module m = open_module(mod);

    puts("/* Autogenerated C header for Gont module " + 
	 m.mod_name + " */\n" +
	 "#ifndef GONT_MODULE_IFACE_" + m.mod_name + "_INCLUDED\n" +
	 "#define GONT_MODULE_IFACE_" + m.mod_name + "_INCLUDED 1\n");

    Hashtbl::iter(m.mod_vars, dump_var);
    Hashtbl::iter(m.mod_tydefs, dump_tysym);
    
    puts("#endif /* GONT_MODULE_IFACE_" + m.mod_name + "_INCLUDED */\n"
	 "/* EOF */\n");
}

// vim:sts=4:sw=4
