/**
+ 64 bit Bolek

We will now teach Bolek to use 64 integers for its 
computations.  Until now, if you were using 32 bit machine,
Bolek operated on 32 bit signed integers.  Gont standard library
provides Int64 module for 64 bit computations, regardless of whatever
machine is currently in use.

Module groups set of functions, variables and types. Module is roughly
eqvialent of .c/.h pair in C.  In Gont there is .g/.gi pair.  .gi stands
for Gont interface.  Interface file defines what types, functions
and variables are visible outside the module.  Module can define local
functions, just by not including them in interface file.  Module can
also not declare internal types or declare them as abstract, so other
modules won't have any access to them. 

.g file contains Gont source. Gont compiler checks whether implementation
conforms interface, i.e. whether all declared functions and variables
are implemented, and wheteher or not they have matching types.

.gi files are compiled into .xgi files, .g files are compiled into .xgo
and .ksi files.  Both .xgi and .xgo are XML files with various information
about modules.  .xgi files for module needs to be visible to the compiler 
whenever some other module references it, as well as when compiling
it's implementation file, so compiler can cross-check it.
As a special exception, if compiler is
passed [-empty-iface] (or [-e], which is the same) switch, it assumes
interface for currently compiled implementation file is empty, i.e.
defines no values nor types.

Modules can contain init and fini sections.  Execution of Gont
program constitutes of running init sections of each linked module
in turn, and then executing fini sections in reverse order.

Libraries, including Gont standard library, are composited of modules.

Modules have alpha-numeric names, starting with uppercase letter.
Name of module is derived from name of source file by stripping off
.g suffix, and capitalizing first letter. So name [foo.g] defines
module [Foo], while [fooBar.g] -- [FooBar].
*/

void process_line()
{
	Int64::t read_x_or_int()		// {0}
	{
		string v = read_string();
		
		if (v == "x" || v == "X")
			return last_expr;	// {1}
		else 
			return Int64::of_string(v);	// {2}
	}

	puts("bolek LXIV > ");

	Int64::t x, y, res;
	string op;
	
	(x, op, y) = (read_x_or_int(), read_string(), read_x_or_int());

	switch op {
			// {3}
	case "+": 	res = Int64::add(x, y);
	case "-": 	res = Int64::sub(x, y);
	case ":":
	case "/": 	res = Int64::div(x, y);
	case "*": 	res = Int64::mul(x, y);
	case _:
		// {4}
		raise User_error("bad operator");
	}
	
	// {5}
	puts("result: " + Int64::to_string(res, 10) + "\n");
	last_expr = res;
}

// {6}
Int64::t last_expr = Int64::zero;	

// {7}
exception string User_error;

section init
{
	try {
		for (;;)
			try {
				process_line();
			} with {
			// {8}
			case Invalid_argument(_):
				puts("bad number\n");
			case User_error(desc):
				puts(desc + "\n");
			}
	} with {
	case End_of_file:
		puts("bye\n");
	}
}

/**
{0} Symbol [foo] from module [Bar] is accessed with [Bar::foo].  It is 
also possible to import all symbols from module [Bar] into current
namespace by using [open Bar;] directive.  However it is not recommended
for standard library modules, as they define lots of symbols that might
clash with your functions.  They are simply designed to be used with
Prefixed::names.

Back to our example [Int64::t] is type of 64 bit integers.  Types declared
in modules can be used where basic type like [int] or [string] could be.
However there is no user-space operator overloading in Gont.

{1} We use [last_expr] here, while it has been defined in much later
in {6}.  It's another feature of Gont -- global symbols in module
can be freerly used before they are defined, no matter whether or not
they are in interface.  Therfore there is no need for prototyping local
functions.  

{2} [Int64::of_string] works much like [atoi] but returns 64 bit number.
Note that we do not catch any exceptions here, they are catched later,
in {7}.

{3} [Int64::add], [Int64::sub] etc. provide usual set of operations on
integers.

{4} In case of invalid operator, instead of printing error message,
as we used to, we raise an exception.  [User_error] is our own, local exception,
defined in {7}, it takes string argument so we supply one.
So when user writes "2 ble 3" [User_error("bad operator")] is raised
and control is transferred outside [process_line()] function, out to
init section, where it is catched and error message is printed.

{5} [Int64::to_string(v, base)] converts [v] to string, using radix 
[base]. 

{6} Here we finally define [last_expr] to be equal to variable 
[Int64::zero].

{7} Here we define our exception.  Note syntax used to specify exception
argument.

{8} Here we catch exceptions from [process_line()] function.
When we get one of [User_error] or [Invalid_argument] exceptions
we print error message and resume loop iterations.  Otherwise
exceptions is raised again, [End_of_file] (which as you can guess
is raised by routines, that ran out of input) is catched one level
up, outside the loop.  So when after catching [End_of_file] "bye"
is printed and execution of program finishes.
*/
